@@ -95,10 +95,10 @@ class InvalidHeader(BadHttpMessage):
 
 
 class BadStatusLine(BadHttpMessage):
-    def __init__(self, line: str = "") -> None:
+    def __init__(self, line: str = "", error: Optional[str] = None) -> None:
         if not isinstance(line, str):
             line = repr(line)
-        super().__init__(f"Bad status line {line!r}")
+        super().__init__(error or f"Bad status line {line!r}")
         self.args = (line,)
         self.line = line
 
